# stdlib
import json
from typing import Dict
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from typing_extensions import final

# syft absolute
from syft import serialize
from syft.core.common.message import ImmediateSyftMessageWithReply
from syft.core.common.message import ImmediateSyftMessageWithoutReply
from syft.core.common.serde.deserialize import _deserialize
from syft.core.common.serde.serializable import bind_protobuf
from syft.core.common.uid import UID
from syft.core.io.address import Address
from syft.proto.grid.messages.dataset_messages_pb2 import (
    CreateDatasetMessage as CreateDatasetMessage_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    CreateDatasetResponse as CreateDatasetResponse_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    DeleteDatasetMessage as DeleteDatasetMessage_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    DeleteDatasetResponse as DeleteDatasetResponse_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    GetDatasetInfoMessage as GetDatasetInfoMessage_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    GetDatasetInfoResponse as GetDatasetInfoResponse_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    GetDatasetMessage as GetDatasetMessage_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    GetDatasetResponse as GetDatasetResponse_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    GetDatasetsInfoMessage as GetDatasetsInfoMessage_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    GetDatasetsInfoResponse as GetDatasetsInfoResponse_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    GetDatasetsMessage as GetDatasetsMessage_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    GetDatasetsResponse as GetDatasetsResponse_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    UpdateDatasetMessage as UpdateDatasetMessage_PB,
)
from syft.proto.grid.messages.dataset_messages_pb2 import (
    UpdateDatasetResponse as UpdateDatasetResponse_PB,
)


@bind_protobuf
@final
class CreateDatasetMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> CreateDatasetMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: CreateDatasetMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return CreateDatasetMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: CreateDatasetMessage_PB,
    ) -> "CreateDatasetMessage":
        """Creates a CreateDatasetMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: CreateDatasetMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return CreateDatasetMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return CreateDatasetMessage_PB


@bind_protobuf
@final
class CreateDatasetResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> CreateDatasetResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return CreateDatasetResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: CreateDatasetResponse_PB,
    ) -> "CreateDatasetResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return CreateDatasetResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return CreateDatasetResponse_PB


@bind_protobuf
@final
class GetDatasetMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> GetDatasetMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetDatasetMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetDatasetMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetDatasetMessage_PB,
    ) -> "GetDatasetMessage":
        """Creates a GetDatasetMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetDatasetMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetDatasetMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetDatasetMessage_PB


@bind_protobuf
@final
class GetDatasetResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> GetDatasetResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetDatasetResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: GetDatasetResponse_PB,
    ) -> "GetDatasetResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetDatasetResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetDatasetResponse_PB


@bind_protobuf
@final
class GetDatasetsMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> GetDatasetsMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetDatasetsMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetDatasetsMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetDatasetsMessage_PB,
    ) -> "GetDatasetsMessage":
        """Creates a GetDatasetsMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetDatasetsMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetDatasetsMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetDatasetsMessage_PB


@bind_protobuf
@final
class GetDatasetsResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> GetDatasetsResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetDatasetsResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: GetDatasetsResponse_PB,
    ) -> "GetDatasetsResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetDatasetsResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetDatasetsResponse_PB


@bind_protobuf
@final
class GetDatasetInfoMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> GetDatasetInfoMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetDatasetInfoMessage_PB
        .. note::
            This method is purely an internal method. Please use object.serialize() or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetDatasetInfoMessage_PB(
            msg_id=self.id.serialize(),
            address=self.address.serialize(),
            content=json.dumps(self.content),
            reply_to=self.reply_to.serialize(),
        )

    @staticmethod
    def _proto2object(
        proto: GetDatasetInfoMessage_PB,
    ) -> "GetDatasetInfoMessage":
        """Creates a GetDatasetInfoMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetDatasetInfoMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetDatasetInfoMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetDatasetInfoMessage_PB


@bind_protobuf
@final
class GetDatasetInfoResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> GetDatasetInfoResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use object.serialize() or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetDatasetInfoResponse_PB(
            msg_id=self.id.serialize(),
            address=self.address.serialize(),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: GetDatasetInfoResponse_PB,
    ) -> "GetDatasetInfoResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetDatasetInfoResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetDatasetInfoResponse_PB


@bind_protobuf
@final
class GetDatasetsInfoMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> GetDatasetsInfoMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetDatasetsInfoMessage_PB
        .. note::
            This method is purely an internal method. Please use object.serialize() or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetDatasetsInfoMessage_PB(
            msg_id=self.id.serialize(),
            address=self.address.serialize(),
            content=json.dumps(self.content),
            reply_to=self.reply_to.serialize(),
        )

    @staticmethod
    def _proto2object(
        proto: GetDatasetsInfoMessage_PB,
    ) -> "GetDatasetsInfoMessage":
        """Creates a GetDatasetsInfoMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: GetDatasetsInfoMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetDatasetsInfoMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetDatasetsInfoMessage_PB


@bind_protobuf
@final
class GetDatasetsInfoResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> GetDatasetsInfoResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use object.serialize() or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetDatasetsInfoResponse_PB(
            msg_id=self.id.serialize(),
            address=self.address.serialize(),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: GetDatasetsInfoResponse_PB,
    ) -> "GetDatasetsInfoResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetDatasetsInfoResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetDatasetsInfoResponse_PB


@bind_protobuf
@final
class UpdateDatasetMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> UpdateDatasetMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: UpdateDatasetMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return UpdateDatasetMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: UpdateDatasetMessage_PB,
    ) -> "UpdateDatasetMessage":
        """Creates a UpdateDatasetMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: UpdateDatasetMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return UpdateDatasetMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return UpdateDatasetMessage_PB


@bind_protobuf
@final
class UpdateDatasetResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> UpdateDatasetResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return UpdateDatasetResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: UpdateDatasetResponse_PB,
    ) -> "UpdateDatasetResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return UpdateDatasetResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return UpdateDatasetResponse_PB


@bind_protobuf
@final
class DeleteDatasetMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        content: Dict,
        reply_to: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.content = content

    def _object2proto(self) -> DeleteDatasetMessage_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: DeleteDatasetMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return DeleteDatasetMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            content=json.dumps(self.content),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: DeleteDatasetMessage_PB,
    ) -> "DeleteDatasetMessage":
        """Creates a DeleteDatasetMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: DeleteDatasetMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return DeleteDatasetMessage(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            content=json.loads(proto.content),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return DeleteDatasetMessage_PB


@bind_protobuf
@final
class DeleteDatasetResponse(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        address: Address,
        status_code: int,
        content: Dict,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.status_code = status_code
        self.content = content

    def _object2proto(self) -> DeleteDatasetResponse_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: SignalingOfferMessage_PB
        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return DeleteDatasetResponse_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            status_code=self.status_code,
            content=json.dumps(self.content),
        )

    @staticmethod
    def _proto2object(
        proto: DeleteDatasetResponse_PB,
    ) -> "DeleteDatasetResponse":
        """Creates a SignalingOfferMessage from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of SignalingOfferMessage
        :rtype: SignalingOfferMessage
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return DeleteDatasetResponse(
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            status_code=proto.status_code,
            content=json.loads(proto.content),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for
        details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return DeleteDatasetResponse_PB



# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# third party
import sklearn
import sklearn.linear_model

# syft relative
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..util import generic_update_ast

LIB_NAME = "sklearn"
PACKAGE_SUPPORT = {"lib": LIB_NAME}


def create_ast(client: TypeAny = None) -> Globals:
    ast = Globals(client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [
        ("sklearn", sklearn),
        ("sklearn.linear_model", sklearn.linear_model),
    ]

    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        ("sklearn.base", "sklearn.base", sklearn.base),
        # linear_model
        # LogisticRegression
        (
            "sklearn.linear_model.LogisticRegression",
            "sklearn.linear_model.LogisticRegression",
            sklearn.linear_model.LogisticRegression,
        ),
        # LinearRegression
        (
            "sklearn.linear_model.LinearRegression",
            "sklearn.linear_model._base.LinearRegression",
            sklearn.linear_model._base.LinearRegression,
        ),
    ]

    methods: TypeList[TypeTuple[str, str]] = [
        # linear_model
        # LogisticRegression
        (
            "sklearn.linear_model.LogisticRegression.fit",
            "sklearn.linear_model.LogisticRegression",
        ),
        ("sklearn.linear_model.LogisticRegression.predict", "pandas.DataFrame"),
        # LinearRegression
        (
            "sklearn.linear_model.LinearRegression.fit",
            "sklearn.linear_model._base.LinearRegression",
        ),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)



# stdlib
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft relative
from ..... import serialize
from .....proto.core.node.common.action.save_object_pb2 import (
    SaveObjectAction as SaveObjectAction_PB,
)
from ....common.serde.deserialize import _deserialize
from ....common.serde.serializable import Serializable
from ....common.serde.serializable import bind_protobuf
from ....common.uid import UID
from ....io.address import Address
from ....store.storeable_object import StorableObject
from ...abstract.node import AbstractNode
from .common import ImmediateActionWithoutReply


@bind_protobuf
class SaveObjectAction(ImmediateActionWithoutReply, Serializable):
    def __init__(
        self,
        obj: StorableObject,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.obj = obj

    def __repr__(self) -> str:
        obj_str = str(self.obj)
        # make obj_str of reasonable length, if too long: cut into begin and end
        neg_index = max(-50, -len(obj_str) + 50)
        obj_str = obj_str = (
            obj_str[:50]
            if len(obj_str) < 50
            else obj_str[:50] + " ... " + obj_str[neg_index:]
        )
        return f"SaveObjectAction {obj_str}"

    def execute_action(self, node: AbstractNode, verify_key: VerifyKey) -> None:
        self.obj.read_permissions = {
            node.verify_key: node.id,
            verify_key: None,  # we dont have the passed in sender's UID
        }
        node.store[self.obj.id] = self.obj

    def _object2proto(self) -> SaveObjectAction_PB:
        obj = self.obj._object2proto()
        addr = serialize(self.address)
        return SaveObjectAction_PB(obj=obj, address=addr)

    @staticmethod
    def _proto2object(proto: SaveObjectAction_PB) -> "SaveObjectAction":
        obj = _deserialize(blob=proto.obj)
        addr = _deserialize(blob=proto.address)
        return SaveObjectAction(obj=obj, address=addr)

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return SaveObjectAction_PB



# stdlib
from collections import namedtuple

# third party
import torch
from torchvision import models


class Vgg16(torch.nn.Module):
    def __init__(self, requires_grad=False):
        super(Vgg16, self).__init__()
        vgg_pretrained_features = models.vgg16(pretrained=True).features
        self.slice1 = torch.nn.Sequential()
        self.slice2 = torch.nn.Sequential()
        self.slice3 = torch.nn.Sequential()
        self.slice4 = torch.nn.Sequential()
        for x in range(4):
            self.slice1.add_module(str(x), vgg_pretrained_features[x])
        for x in range(4, 9):
            self.slice2.add_module(str(x), vgg_pretrained_features[x])
        for x in range(9, 16):
            self.slice3.add_module(str(x), vgg_pretrained_features[x])
        for x in range(16, 23):
            self.slice4.add_module(str(x), vgg_pretrained_features[x])
        if not requires_grad:
            for param in self.parameters():
                param.requires_grad = False

    def forward(self, X):
        h = self.slice1(X)
        h_relu1_2 = h
        h = self.slice2(h)
        h_relu2_2 = h
        h = self.slice3(h)
        h_relu3_3 = h
        h = self.slice4(h)
        h_relu4_3 = h
        vgg_outputs = namedtuple(
            "VggOutputs", ["relu1_2", "relu2_2", "relu3_3", "relu4_3"]
        )
        out = vgg_outputs(h_relu1_2, h_relu2_2, h_relu3_3, h_relu4_3)
        return out



# third party
import PIL
import numpy as np
import torch
import torchvision

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...lib.torch.tensor_util import protobuf_tensor_deserializer
from ...lib.torch.tensor_util import protobuf_tensor_serializer
from ...proto.lib.torch.tensor_pb2 import TensorData


def object2proto(obj: PIL.Image.Image) -> TensorData:
    image_tensor = torch.Tensor(np.array(obj))
    tensor_proto = protobuf_tensor_serializer(image_tensor)

    return tensor_proto


def proto2object(proto: TensorData) -> PIL.Image.Image:
    image_tensor = protobuf_tensor_deserializer(proto)
    if image_tensor.dim() == 3:
        image_tensor = image_tensor.permute(2, 0, 1)
    image_obj = torchvision.transforms.functional.to_pil_image(image_tensor)
    return image_obj


GenerateWrapper(
    wrapped_type=PIL.Image.Image,
    import_path="PIL.Image.Image",
    protobuf_scheme=TensorData,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# third party
import torch as th

# syft absolute
import syft as sy
from syft import serialize
from syft.core.common.uid import UID
from syft.core.io.address import Address
from syft.core.io.location import SpecificLocation
from syft.core.node.common.action.save_object_action import SaveObjectAction
from syft.core.store.storeable_object import StorableObject


def test_save_object_action_serde() -> None:
    obj = th.tensor([1, 2, 3])
    addr = Address(network=SpecificLocation(), device=SpecificLocation())

    storable = StorableObject(id=UID(), data=obj)
    msg = SaveObjectAction(obj=storable, address=addr)

    blob = serialize(msg)

    msg2 = sy.deserialize(blob=blob)

    assert (msg2.obj.data == msg.obj.data).all()

    # Tensors do not automatically get IDs anymore
    # assert msg2.obj.id == msg.obj.id
    assert msg2.address == msg.address



# stdlib
from typing import List
from typing import Optional

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey

# syft relative
from ...common.message import SignedImmediateSyftMessageWithoutReply
from ...io.location import Location
from ...io.location import SpecificLocation
from ..common.action.common import Action
from .vm import VirtualMachine


class PlanVirtualMachine(VirtualMachine):  # type: ignore
    def __init__(
        self,
        *,  # Trasterisk
        name: Optional[str] = None,
        network: Optional[Location] = None,
        domain: Optional[Location] = None,
        device: Optional[Location] = None,
        vm: SpecificLocation = SpecificLocation(),
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
    ):
        super().__init__(
            name=name,
            network=network,
            domain=domain,
            device=device,
            vm=vm,
            signing_key=signing_key,
            verify_key=verify_key,
        )
        self.record = False
        self.recorded_actions: List[Action] = []

    def recv_immediate_msg_without_reply(
        self, msg: SignedImmediateSyftMessageWithoutReply
    ) -> None:
        if self.record and isinstance(msg.message, Action):
            self.recorded_actions.append(msg.message)
        super().recv_immediate_msg_without_reply(msg)

    def record_actions(self) -> None:
        self.record = True

    def stop_recording(self) -> None:
        self.record = False



# For creating Duet tests from notebooks, adapt your notebooks to the following rules:
# 1. The DO/DS notebooks have a predefined format:
#  <test_name>_Data_Owner.ipynb and <test_name>_Data_Scientist.ipynb.
#
# 2. Each notebook must have special markdown cells called Checkpoints, with the
#    following properties:
#     - The cell begins with the text "Checkpoint <int>: "
#       used by the script for extracting the step number.
#     - Each Checkpoint cell acts as a barrier in the generated scripts.
#     - In the DO tests, getting to Checkpoint N means creating a file to mark the
#       checkpoint and waiting for the DS instance to get the Checkpoint N as well.
#     - In the DS tests, getting to Checkpoint N means creating a file to mark the
#       checkpoint and waiting for the DO instance to get to Checkpoint N + 1.
#     - The DS notebook instance should always end with a checkpoint cell, to wait for
#       the DO instance to assert the results.

# stdlib
from collections import defaultdict
import json
import os
from pathlib import Path
import re
import secrets
import shutil
from typing import Dict

# third party
from nbconvert import PythonExporter
from nbconvert.writers import FilesWriter
import nbformat

tests = defaultdict(list)
output_dir = Path("tests/syft/notebooks")
checkpoint_dir = Path("tests/syft/notebooks/checkpoints")

SLEEP_TIME = 360
PORTS: Dict[str, int] = {}

try:
    os.makedirs(output_dir, exist_ok=True)
except BaseException as e:
    print("os.makedirs failed ", e)

try:
    shutil.rmtree(checkpoint_dir)
except BaseException as e:
    print("rmtree failed ", e)

try:
    os.makedirs(checkpoint_dir, exist_ok=True)
except BaseException as e:
    print("os.makedirs failed ", e)

testcase_lib = {}

for path in (
    list(Path("examples/homomorphic-encryption").rglob("*.ipynb"))
    + list(Path("examples/duet/dcgan").rglob("*.ipynb"))
    + list(Path("examples/duet/super_resolution").rglob("*.ipynb"))
    + list(Path("examples/private-ai-series/duet_basics").rglob("*.ipynb"))
    + list(Path("examples/private-ai-series/duet_iris_classifier").rglob("*.ipynb"))
    + list(Path("examples/differential-privacy/opacus").rglob("*.ipynb"))
    + list(Path("examples/duet/mnist").rglob("*.ipynb"))
    + list(Path("examples/duet/mnist_lightning").rglob("*.ipynb"))
):
    if ".ipynb_checkpoints" in str(path):
        continue

    testname = re.sub("[^0-9a-zA-Z]+", "_", str(path))
    output = output_dir / testname

    file_name = str(path.stem)
    is_do = False
    is_ds = False

    if file_name.endswith("_Data_Scientist"):
        testcase = file_name.replace("_Data_Scientist", "")
        tests[testcase].append(testname)
        is_ds = True
    elif file_name.endswith("_Data_Owner"):
        testcase = file_name.replace("_Data_Owner", "")
        tests[testcase].append(testname)
        is_do = True
    else:
        continue

    # generate a unique port for this testcase
    while testcase not in PORTS:
        PORT = secrets.choice(range(21000, 22000))
        if PORT not in PORTS.values():
            # unique port so set
            PORTS[testcase] = PORT
    load_lib_search = r"load_lib\(\W+([a-z_-]+)\W+\)"

    with open(path, "r") as f:
        load_lib_results = re.search(load_lib_search, str(f.read()), re.IGNORECASE)
        if load_lib_results:
            lib_name = load_lib_results.group(1)
            testcase_lib[testcase] = lib_name

    notebook_nodes = nbformat.read(path, as_version=4)

    custom_cell = nbformat.v4.new_code_cell(
        source="""
import os
import time
import asyncio
from pathlib import Path

loop = asyncio.get_event_loop()
"""
    )
    notebook_nodes["cells"].insert(0, custom_cell)

    for idx, cell in enumerate(notebook_nodes["cells"]):
        if cell["cell_type"] == "code" and "loopback=True" in cell["source"]:
            network_url = f"http://127.0.0.1:{PORTS[testcase]}"
            notebook_nodes["cells"][idx]["source"] = cell["source"].replace(
                "loopback=True", f"loopback=True, network_url='{network_url}'"
            )
        if cell["cell_type"] == "markdown" and "Checkpoint" in cell["source"]:
            checkpoint = (
                cell["source"]
                .lower()
                .split("checkpoint")[1]
                .strip()
                .split(":")[0]
                .strip()
            )

            testcase_checkpoint_dir = f"checkpoints/{testcase}/"

            # For DO, we wait until DS gets to the same checkpoint
            if is_do:
                ck_file = testcase_checkpoint_dir + (
                    testcase + "_DO_checkpoint_" + str(checkpoint)
                )
                wait_file = testcase_checkpoint_dir + (
                    testcase + "_DS_checkpoint_" + str(checkpoint)
                )
                checkpoint_cell = nbformat.v4.new_code_cell(
                    source=f"""
Path(\"{ck_file}\").touch()
for retry in range({SLEEP_TIME}):
    if Path(\"{wait_file}\").exists():
        break
    task = loop.create_task(asyncio.sleep(1))
    loop.run_until_complete(task)
                                                            """
                )

            # For DS, we wait until DO gets to the next checkpoint
            elif is_ds:
                ck_file = testcase_checkpoint_dir + (
                    testcase + "_DS_checkpoint_" + str(checkpoint)
                )
                wait_file = testcase_checkpoint_dir + (
                    testcase + "_DO_checkpoint_" + str(int(checkpoint) + 1)
                )
                checkpoint_cell = nbformat.v4.new_code_cell(
                    source=f"""
Path(\"{ck_file}\").touch()
for retry in range({SLEEP_TIME}):
    if Path(\"{wait_file}\").exists():
        break
    task = loop.create_task(asyncio.sleep(1))
    loop.run_until_complete(task)
assert Path(\"{wait_file}\").exists()
                                                            """
                )
            notebook_nodes["cells"][idx] = checkpoint_cell

    try:
        exporter = PythonExporter()

        (body, resources) = exporter.from_notebook_node(notebook_nodes)
        write_file = FilesWriter()

        # replace empty cells with print statements for easy debugging
        empty_cell = "# In[ ]:"
        counter = 1
        cell_type = "DO" if is_do else "DS"
        while empty_cell in body:
            body = body.replace(empty_cell, f"print('{cell_type} Cell: {counter}')", 1)
            counter += 1

        # replace any test variables / lines to make things faster in test mode
        json_file = f"{path}.json"
        if os.path.exists(json_file):
            with open(json_file, "r") as f:
                json_rules = json.loads(f.read())

                for rules in json_rules["replace_lines"]:
                    try:
                        body = re.sub(
                            rules["match"], rules["replace"], body, flags=re.MULTILINE
                        )
                    except Exception as e:
                        print(f"Failed to replace rule {rules} for test: {path}. {e}")

        write_file.write(output=body, resources=resources, notebook_name=str(output))
    except Exception as e:
        print(f"There was a problem exporting the file(s): {e}")

for case in tests:
    test = tests[case]
    if len(test) != 2:
        print("invalid testcase ", test)

    print(case, test)

    template = open(output_dir / "duet_test.py.template").read()

    output_py = template.replace("{{TESTCASE}}", str(case))
    output_py = output_py.replace("21000", f"{PORTS[case]}")
    output_py = output_py.replace("checkpoints", f"checkpoints/{case}")

    for script in test:
        if "Data_Owner" in script:
            output_py = output_py.replace("{{DO_SCRIPT}}", script)
        elif "Data_Scientist" in script:
            output_py = output_py.replace("{{DS_SCRIPT}}", script)

    decorator = ""
    if case in testcase_lib:
        lib_name = testcase_lib[case]
        decorator = f"@pytest.mark.vendor(lib='{lib_name}')"
    output_py = output_py.replace("{{LIB_DECORATOR}}", decorator)

    with open(output_dir / f"duet_{case}_test.py", "w") as out_py:
        out_py.write(output_py)



# stdlib
from typing import Any
from typing import Callable as CallableT

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft

# syft relative
from .core.common.serde.serializable import Serializable
from .core.common.serde.serializable import bind_protobuf
from .util import aggressive_set_attr

module_type = type(syft)


# this will overwrite the ._sy_serializable_wrapper_type with an auto generated
# wrapper which will basically just hold the object being wrapped.
def GenerateWrapper(
    wrapped_type: type,
    import_path: str,
    protobuf_scheme: GeneratedProtocolMessageType,
    type_object2proto: CallableT,
    type_proto2object: CallableT,
) -> None:
    @bind_protobuf
    class Wrapper(Serializable):
        def __init__(self, value: object):
            self.obj = value

        def _object2proto(self) -> Any:
            return type_object2proto(self.obj)

        @staticmethod
        def _proto2object(proto: Any) -> Any:
            return type_proto2object(proto)

        @staticmethod
        def get_protobuf_schema() -> GeneratedProtocolMessageType:
            return protobuf_scheme

        def upcast(self) -> Any:
            return self.obj

    # set __module__ and __name__
    module_parts = import_path.split(".")
    klass = module_parts.pop()
    Wrapper.__name__ = f"{klass}Wrapper"
    Wrapper.__module__ = f"syft.wrappers.{'.'.join(module_parts)}"
    # create a fake module `wrappers` under `syft`
    if "wrappers" not in syft.__dict__:
        syft.__dict__["wrappers"] = module_type(name="wrappers")
    # for each part of the path, create a fake module and add it to it's parent
    parent = syft.__dict__["wrappers"]
    for n in module_parts:
        if n not in parent.__dict__:
            parent.__dict__[n] = module_type(name=n)
        parent = parent.__dict__[n]
    # finally add our wrapper class to the end of the path
    parent.__dict__[Wrapper.__name__] = Wrapper

    aggressive_set_attr(
        obj=wrapped_type, name="_sy_serializable_wrapper_type", attr=Wrapper
    )


def GenerateProtobufWrapper(
    cls_pb: GeneratedProtocolMessageType, import_path: str
) -> None:
    def object2proto(obj: Any) -> Any:
        return obj

    def proto2object(proto: Any) -> Any:
        return proto

    GenerateWrapper(
        wrapped_type=cls_pb,
        import_path=import_path,
        protobuf_scheme=cls_pb,
        type_object2proto=object2proto,
        type_proto2object=proto2object,
    )



# third party
import torch as th

# syft absolute
import syft as sy
from syft.core.node.common.service.obj_search_permission_service import (
    ImmediateObjectSearchPermissionUpdateService,
)
from syft.core.node.common.service.obj_search_permission_service import (
    ObjectSearchPermissionUpdateMessage,
)


def test_object_search_permissons_update_message_serde() -> None:
    bob_phone = sy.Device(name="Bob's iPhone")
    bob_phone_client = bob_phone.get_client()

    ptr = th.tensor([1, 2, 3]).send(bob_phone_client)

    msg = ObjectSearchPermissionUpdateMessage(
        add_instead_of_remove=True,
        target_verify_key=bob_phone_client.verify_key,
        target_object_id=ptr.id_at_location,
        address=bob_phone_client.address,
    )

    blob = sy.serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == msg2.address
    assert msg.add_instead_of_remove == msg2.add_instead_of_remove
    assert msg.target_verify_key == msg2.target_verify_key
    assert msg.target_object_id == msg2.target_object_id


def test_object_search_permissons_update_execute_add() -> None:
    bob_phone = sy.Device(name="Bob's iPhone")
    bob_phone_client = bob_phone.get_client()

    ptr = th.tensor([1, 2, 3]).send(bob_phone_client)

    msg = ObjectSearchPermissionUpdateMessage(
        add_instead_of_remove=True,
        target_verify_key=bob_phone_client.verify_key,
        target_object_id=ptr.id_at_location,
        address=bob_phone_client.address,
    )

    ImmediateObjectSearchPermissionUpdateService.process(
        node=bob_phone, msg=msg, verify_key=bob_phone.verify_key
    )

    assert (
        bob_phone.store[ptr.id_at_location].search_permissions[
            bob_phone_client.verify_key
        ]
        == msg.id
    )


def test_object_search_permissons_update_execute_remove() -> None:
    bob_phone = sy.Device(name="Bob's iPhone")
    bob_phone_client = bob_phone.get_client()

    ptr = th.tensor([1, 2, 3]).send(bob_phone_client)

    msg = ObjectSearchPermissionUpdateMessage(
        add_instead_of_remove=False,
        target_verify_key=bob_phone_client.verify_key,
        target_object_id=ptr.id_at_location,
        address=bob_phone_client.address,
    )

    bob_phone.store[ptr.id_at_location].search_permissions[bob_phone.verify_key] = None

    ImmediateObjectSearchPermissionUpdateService.process(
        node=bob_phone, msg=msg, verify_key=bob_phone.verify_key
    )

    assert (
        bob_phone_client.verify_key
        not in bob_phone.store[ptr.id_at_location].search_permissions
    )



# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# syft relative
from . import session  # noqa: 401
from . import share  # noqa: 401
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..util import generic_update_ast

LIB_NAME = "sympc"
PACKAGE_SUPPORT = {
    "lib": LIB_NAME,
    "torch": {"min_version": "1.6.0", "max_version": "1.8.0"},
    "python": {"min_version": (3, 7), "max_version": (3, 9, 99)},
}


def create_ast(client: TypeAny = None) -> Globals:
    # third party
    import sympc

    # syft relative
    from . import session  # noqa: 401
    from . import share  # noqa: 401

    ast = Globals(client=client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [
        ("sympc", sympc),
        ("sympc.session", sympc.session),
        ("sympc.tensor", sympc.tensor),
        ("sympc.protocol", sympc.protocol),
        ("sympc.store", sympc.store),
        ("sympc.protocol.fss", sympc.protocol.fss),
        ("sympc.protocol.fss.fss", sympc.protocol.fss.fss),
        ("sympc.protocol.spdz", sympc.protocol.spdz),
        ("sympc.protocol.spdz.spdz", sympc.protocol.spdz.spdz),
        ("sympc.utils", sympc.utils),
    ]

    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        ("sympc.session.Session", "sympc.session.Session", sympc.session.Session),
        ("sympc.store.CryptoStore", "sympc.store.CryptoStore", sympc.store.CryptoStore),
        (
            "sympc.tensor.ShareTensor",
            "sympc.tensor.ShareTensor",
            sympc.tensor.ShareTensor,
        ),
    ]

    methods: TypeList[TypeTuple[str, str]] = [
        ("sympc.store.CryptoStore.get_primitives_from_store", "syft.lib.python.List"),
        ("sympc.session.Session.crypto_store", "sympc.store.CryptoStore"),
        ("sympc.protocol.fss.fss.mask_builder", "sympc.tensor.ShareTensor"),
        ("sympc.protocol.fss.fss.evaluate", "sympc.tensor.ShareTensor"),
        ("sympc.protocol.spdz.spdz.mul_parties", "sympc.tensor.ShareTensor"),
        ("sympc.protocol.spdz.spdz.spdz_mask", "syft.lib.python.Tuple"),
        ("sympc.protocol.spdz.spdz.div_wraps", "sympc.tensor.ShareTensor"),
        (
            "sympc.session.Session.przs_generate_random_share",
            "sympc.tensor.ShareTensor",
        ),
        (
            "sympc.store.CryptoStore.populate_store",
            "syft.lib.python._SyNone",
        ),
        (
            "sympc.utils.get_new_generator",
            "torch.Generator",
        ),
        (
            "sympc.tensor.ShareTensor.__add__",
            "sympc.tensor.ShareTensor",
        ),
        (
            "sympc.tensor.ShareTensor.__sub__",
            "sympc.tensor.ShareTensor",
        ),
        (
            "sympc.tensor.ShareTensor.__rmul__",
            "sympc.tensor.ShareTensor",
        ),
        (
            "sympc.tensor.ShareTensor.__mul__",
            "sympc.tensor.ShareTensor",
        ),
        (
            "sympc.tensor.ShareTensor.__matmul__",
            "sympc.tensor.ShareTensor",
        ),
        (
            "sympc.tensor.ShareTensor.__truediv__",
            "sympc.tensor.ShareTensor",
        ),
        (
            "sympc.tensor.ShareTensor.__rmatmul__",
            "sympc.tensor.ShareTensor",
        ),
        (
            "sympc.tensor.ShareTensor.numel",
            "syft.lib.python.Int",  # FIXME: Can't we just return an int??
        ),
        (
            "sympc.tensor.ShareTensor.T",
            "sympc.tensor.ShareTensor",
        ),
        ("sympc.tensor.ShareTensor.unsqueeze", "sympc.tensor.ShareTensor"),
        ("sympc.tensor.ShareTensor.view", "sympc.tensor.ShareTensor"),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/service/signaling_service.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.core.node.common import (
    metadata_pb2 as proto_dot_core_dot_node_dot_common_dot_metadata__pb2,
)

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/grid/service/signaling_service.proto",
    package="syft.grid.service",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n*proto/grid/service/signaling_service.proto\x12\x11syft.grid.service\x1a%proto/core/common/common_object.proto\x1a%proto/core/node/common/metadata.proto\x1a\x1bproto/core/io/address.proto"\x90\x01\n\x16RegisterNewPeerMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address"|\n\x1aPeerSuccessfullyRegistered\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07peer_id\x18\x03 \x01(\t"\xd8\x01\n\x16SignalingAnswerMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07payload\x18\x03 \x01(\t\x12\x36\n\rhost_metadata\x18\x04 \x01(\x0b\x32\x1f.syft.core.node.common.Metadata\x12\x13\n\x0btarget_peer\x18\x05 \x01(\t\x12\x11\n\thost_peer\x18\x06 \x01(\t"\xd7\x01\n\x15SignalingOfferMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07payload\x18\x03 \x01(\t\x12\x36\n\rhost_metadata\x18\x04 \x01(\x0b\x32\x1f.syft.core.node.common.Metadata\x12\x13\n\x0btarget_peer\x18\x05 \x01(\t\x12\x11\n\thost_peer\x18\x06 \x01(\t"\xb9\x01\n\x17OfferPullRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x13\n\x0btarget_peer\x18\x04 \x01(\t\x12\x11\n\thost_peer\x18\x05 \x01(\t"\xba\x01\n\x18\x41nswerPullRequestMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x13\n\x0btarget_peer\x18\x04 \x01(\t\x12\x11\n\thost_peer\x18\x05 \x01(\t"j\n\x19SignalingRequestsNotFound\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address"g\n\x16InvalidLoopBackRequest\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address"g\n\x16\x43loseConnectionMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3',
    dependencies=[
        proto_dot_core_dot_common_dot_common__object__pb2.DESCRIPTOR,
        proto_dot_core_dot_node_dot_common_dot_metadata__pb2.DESCRIPTOR,
        proto_dot_core_dot_io_dot_address__pb2.DESCRIPTOR,
    ],
)


_REGISTERNEWPEERMESSAGE = _descriptor.Descriptor(
    name="RegisterNewPeerMessage",
    full_name="syft.grid.service.RegisterNewPeerMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.service.RegisterNewPeerMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.service.RegisterNewPeerMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.service.RegisterNewPeerMessage.reply_to",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=173,
    serialized_end=317,
)


_PEERSUCCESSFULLYREGISTERED = _descriptor.Descriptor(
    name="PeerSuccessfullyRegistered",
    full_name="syft.grid.service.PeerSuccessfullyRegistered",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.service.PeerSuccessfullyRegistered.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.service.PeerSuccessfullyRegistered.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="peer_id",
            full_name="syft.grid.service.PeerSuccessfullyRegistered.peer_id",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=319,
    serialized_end=443,
)


_SIGNALINGANSWERMESSAGE = _descriptor.Descriptor(
    name="SignalingAnswerMessage",
    full_name="syft.grid.service.SignalingAnswerMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.service.SignalingAnswerMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.service.SignalingAnswerMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="payload",
            full_name="syft.grid.service.SignalingAnswerMessage.payload",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="host_metadata",
            full_name="syft.grid.service.SignalingAnswerMessage.host_metadata",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="target_peer",
            full_name="syft.grid.service.SignalingAnswerMessage.target_peer",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="host_peer",
            full_name="syft.grid.service.SignalingAnswerMessage.host_peer",
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=446,
    serialized_end=662,
)


_SIGNALINGOFFERMESSAGE = _descriptor.Descriptor(
    name="SignalingOfferMessage",
    full_name="syft.grid.service.SignalingOfferMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.service.SignalingOfferMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.service.SignalingOfferMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="payload",
            full_name="syft.grid.service.SignalingOfferMessage.payload",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="host_metadata",
            full_name="syft.grid.service.SignalingOfferMessage.host_metadata",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="target_peer",
            full_name="syft.grid.service.SignalingOfferMessage.target_peer",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="host_peer",
            full_name="syft.grid.service.SignalingOfferMessage.host_peer",
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=665,
    serialized_end=880,
)


_OFFERPULLREQUESTMESSAGE = _descriptor.Descriptor(
    name="OfferPullRequestMessage",
    full_name="syft.grid.service.OfferPullRequestMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.service.OfferPullRequestMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.service.OfferPullRequestMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.service.OfferPullRequestMessage.reply_to",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="target_peer",
            full_name="syft.grid.service.OfferPullRequestMessage.target_peer",
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="host_peer",
            full_name="syft.grid.service.OfferPullRequestMessage.host_peer",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=883,
    serialized_end=1068,
)


_ANSWERPULLREQUESTMESSAGE = _descriptor.Descriptor(
    name="AnswerPullRequestMessage",
    full_name="syft.grid.service.AnswerPullRequestMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.service.AnswerPullRequestMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.service.AnswerPullRequestMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.grid.service.AnswerPullRequestMessage.reply_to",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="target_peer",
            full_name="syft.grid.service.AnswerPullRequestMessage.target_peer",
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="host_peer",
            full_name="syft.grid.service.AnswerPullRequestMessage.host_peer",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1071,
    serialized_end=1257,
)


_SIGNALINGREQUESTSNOTFOUND = _descriptor.Descriptor(
    name="SignalingRequestsNotFound",
    full_name="syft.grid.service.SignalingRequestsNotFound",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.service.SignalingRequestsNotFound.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.service.SignalingRequestsNotFound.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1259,
    serialized_end=1365,
)


_INVALIDLOOPBACKREQUEST = _descriptor.Descriptor(
    name="InvalidLoopBackRequest",
    full_name="syft.grid.service.InvalidLoopBackRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.service.InvalidLoopBackRequest.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.service.InvalidLoopBackRequest.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1367,
    serialized_end=1470,
)


_CLOSECONNECTIONMESSAGE = _descriptor.Descriptor(
    name="CloseConnectionMessage",
    full_name="syft.grid.service.CloseConnectionMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.grid.service.CloseConnectionMessage.msg_id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.grid.service.CloseConnectionMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1472,
    serialized_end=1575,
)

_REGISTERNEWPEERMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_REGISTERNEWPEERMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_REGISTERNEWPEERMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_PEERSUCCESSFULLYREGISTERED.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_PEERSUCCESSFULLYREGISTERED.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_SIGNALINGANSWERMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_SIGNALINGANSWERMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_SIGNALINGANSWERMESSAGE.fields_by_name[
    "host_metadata"
].message_type = proto_dot_core_dot_node_dot_common_dot_metadata__pb2._METADATA
_SIGNALINGOFFERMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_SIGNALINGOFFERMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_SIGNALINGOFFERMESSAGE.fields_by_name[
    "host_metadata"
].message_type = proto_dot_core_dot_node_dot_common_dot_metadata__pb2._METADATA
_OFFERPULLREQUESTMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_OFFERPULLREQUESTMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_OFFERPULLREQUESTMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_ANSWERPULLREQUESTMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_ANSWERPULLREQUESTMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_ANSWERPULLREQUESTMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_SIGNALINGREQUESTSNOTFOUND.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_SIGNALINGREQUESTSNOTFOUND.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_INVALIDLOOPBACKREQUEST.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_INVALIDLOOPBACKREQUEST.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_CLOSECONNECTIONMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_CLOSECONNECTIONMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
DESCRIPTOR.message_types_by_name["RegisterNewPeerMessage"] = _REGISTERNEWPEERMESSAGE
DESCRIPTOR.message_types_by_name[
    "PeerSuccessfullyRegistered"
] = _PEERSUCCESSFULLYREGISTERED
DESCRIPTOR.message_types_by_name["SignalingAnswerMessage"] = _SIGNALINGANSWERMESSAGE
DESCRIPTOR.message_types_by_name["SignalingOfferMessage"] = _SIGNALINGOFFERMESSAGE
DESCRIPTOR.message_types_by_name["OfferPullRequestMessage"] = _OFFERPULLREQUESTMESSAGE
DESCRIPTOR.message_types_by_name["AnswerPullRequestMessage"] = _ANSWERPULLREQUESTMESSAGE
DESCRIPTOR.message_types_by_name[
    "SignalingRequestsNotFound"
] = _SIGNALINGREQUESTSNOTFOUND
DESCRIPTOR.message_types_by_name["InvalidLoopBackRequest"] = _INVALIDLOOPBACKREQUEST
DESCRIPTOR.message_types_by_name["CloseConnectionMessage"] = _CLOSECONNECTIONMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RegisterNewPeerMessage = _reflection.GeneratedProtocolMessageType(
    "RegisterNewPeerMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _REGISTERNEWPEERMESSAGE,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.RegisterNewPeerMessage)
    },
)
_sym_db.RegisterMessage(RegisterNewPeerMessage)

PeerSuccessfullyRegistered = _reflection.GeneratedProtocolMessageType(
    "PeerSuccessfullyRegistered",
    (_message.Message,),
    {
        "DESCRIPTOR": _PEERSUCCESSFULLYREGISTERED,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.PeerSuccessfullyRegistered)
    },
)
_sym_db.RegisterMessage(PeerSuccessfullyRegistered)

SignalingAnswerMessage = _reflection.GeneratedProtocolMessageType(
    "SignalingAnswerMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _SIGNALINGANSWERMESSAGE,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.SignalingAnswerMessage)
    },
)
_sym_db.RegisterMessage(SignalingAnswerMessage)

SignalingOfferMessage = _reflection.GeneratedProtocolMessageType(
    "SignalingOfferMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _SIGNALINGOFFERMESSAGE,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.SignalingOfferMessage)
    },
)
_sym_db.RegisterMessage(SignalingOfferMessage)

OfferPullRequestMessage = _reflection.GeneratedProtocolMessageType(
    "OfferPullRequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _OFFERPULLREQUESTMESSAGE,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.OfferPullRequestMessage)
    },
)
_sym_db.RegisterMessage(OfferPullRequestMessage)

AnswerPullRequestMessage = _reflection.GeneratedProtocolMessageType(
    "AnswerPullRequestMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _ANSWERPULLREQUESTMESSAGE,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.AnswerPullRequestMessage)
    },
)
_sym_db.RegisterMessage(AnswerPullRequestMessage)

SignalingRequestsNotFound = _reflection.GeneratedProtocolMessageType(
    "SignalingRequestsNotFound",
    (_message.Message,),
    {
        "DESCRIPTOR": _SIGNALINGREQUESTSNOTFOUND,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.SignalingRequestsNotFound)
    },
)
_sym_db.RegisterMessage(SignalingRequestsNotFound)

InvalidLoopBackRequest = _reflection.GeneratedProtocolMessageType(
    "InvalidLoopBackRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _INVALIDLOOPBACKREQUEST,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.InvalidLoopBackRequest)
    },
)
_sym_db.RegisterMessage(InvalidLoopBackRequest)

CloseConnectionMessage = _reflection.GeneratedProtocolMessageType(
    "CloseConnectionMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _CLOSECONNECTIONMESSAGE,
        "__module__": "proto.grid.service.signaling_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.grid.service.CloseConnectionMessage)
    },
)
_sym_db.RegisterMessage(CloseConnectionMessage)


# @@protoc_insertion_point(module_scope)



# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
import pytest

# syft absolute
import syft as sy
from syft.core.node.common.service.auth import AuthorizationException
from syft.core.node.common.service.auth import service_auth


def test_service_auth_root_fails() -> None:
    node = sy.Device()
    msg = sy.ReprMessage(address=node.address)

    random_signing_key = SigningKey.generate()
    random_verify_key = random_signing_key.verify_key

    # root_only
    @service_auth(root_only=True)
    def process(node: sy.Device, msg: sy.ReprMessage, verify_key: VerifyKey) -> None:
        pass

    process(node=node, msg=msg, verify_key=node.root_verify_key)

    with pytest.raises(
        AuthorizationException, match="You are not Authorized to access this service"
    ):
        process(node=node, msg=msg, verify_key=random_verify_key)


def test_service_auth_existing_user() -> None:
    node = sy.Device()
    msg = sy.ReprMessage(address=node.address)
    random_signing_key = SigningKey.generate()
    random_verify_key = random_signing_key.verify_key

    # existing_users_only
    @service_auth(existing_users_only=True)
    def process(node: sy.Device, msg: sy.ReprMessage, verify_key: VerifyKey) -> None:
        pass

    with pytest.raises(AuthorizationException, match="User not known."):
        process(node=node, msg=msg, verify_key=random_verify_key)

    # NOTE didn't find a method to add a key to guest_verify_key_registry
    node.guest_verify_key_registry.add(random_verify_key)
    process(node=node, msg=msg, verify_key=random_verify_key)


def test_service_auth_guests_fails() -> None:
    node = sy.Device()
    msg = sy.ReprMessage(address=node.address)
    new_signing_key = SigningKey.generate()
    new_verify_key = new_signing_key.verify_key

    # guests_welcome
    @service_auth(guests_welcome=True)
    def process(node: sy.Device, msg: sy.ReprMessage, verify_key: VerifyKey) -> None:
        pass

    process(node=node, msg=msg, verify_key=new_verify_key)

    assert new_verify_key not in node.guest_verify_key_registry


def test_service_auth_guests_succeeds() -> None:
    node = sy.Device()
    msg = sy.ReprMessage(address=node.address)
    new_signing_key = SigningKey.generate()
    new_verify_key = new_signing_key.verify_key

    # register_new_guests
    @service_auth(guests_welcome=True, register_new_guests=True)
    def process(node: sy.Device, msg: sy.ReprMessage, verify_key: VerifyKey) -> None:
        pass

    process(node=node, msg=msg, verify_key=new_verify_key)

    assert new_verify_key in node.guest_verify_key_registry


def test_service_auth_admin_fails() -> None:
    node = sy.Device()
    msg = sy.ReprMessage(address=node.address)

    random_signing_key = SigningKey.generate()
    random_verify_key = random_signing_key.verify_key

    # Administrator only
    @service_auth(admin_only=True)
    def process(node: sy.Device, msg: sy.ReprMessage, verify_key: VerifyKey) -> None:
        pass

    with pytest.raises(
        AuthorizationException, match="User lacks Administrator credentials."
    ):
        process(node=node, msg=msg, verify_key=random_verify_key)


def test_service_auth_admin_success() -> None:
    node = sy.Device()
    msg = sy.ReprMessage(address=node.address)

    random_signing_key = SigningKey.generate()
    random_verify_key = random_signing_key.verify_key

    # Administrator only
    @service_auth(admin_only=True)
    def process(node: sy.Device, msg: sy.ReprMessage, verify_key: VerifyKey) -> None:
        pass

    # NOTE didn't find a method to add a key to admin_verify_key_registry
    node.admin_verify_key_registry.add(random_verify_key)
    process(node=node, msg=msg, verify_key=random_verify_key)


def test_service_auth_cpl_ofcr_fails() -> None:
    node = sy.Device()
    msg = sy.ReprMessage(address=node.address)

    random_signing_key = SigningKey.generate()
    random_verify_key = random_signing_key.verify_key

    # Compliance Officer only
    @service_auth(cpl_ofcr_only=True)
    def process(node: sy.Device, msg: sy.ReprMessage, verify_key: VerifyKey) -> None:
        pass

    with pytest.raises(
        AuthorizationException, match="User lacks Compliance Officer credentials."
    ):
        process(node=node, msg=msg, verify_key=random_verify_key)


def test_service_auth_cpl_ofcr_success() -> None:
    node = sy.Device()
    msg = sy.ReprMessage(address=node.address)

    random_signing_key = SigningKey.generate()
    random_verify_key = random_signing_key.verify_key

    # Compliance Officer only
    @service_auth(cpl_ofcr_only=True)
    def process(node: sy.Device, msg: sy.ReprMessage, verify_key: VerifyKey) -> None:
        pass

    # NOTE didn't find a method to add a key to cpl_ofcr_verify_key_registry
    node.cpl_ofcr_verify_key_registry.add(random_verify_key)
    process(node=node, msg=msg, verify_key=random_verify_key)


def test_service_auth_admin_success_as_root() -> None:
    node = sy.Device()
    msg = sy.ReprMessage(address=node.address)

    # Administrator only
    @service_auth(admin_only=True)
    def process(node: sy.Device, msg: sy.ReprMessage, verify_key: VerifyKey) -> None:
        pass

    process(node=node, msg=msg, verify_key=node.root_verify_key)


def test_service_auth_cpl_ofcr_success_as_root() -> None:
    node = sy.Device()
    msg = sy.ReprMessage(address=node.address)

    # Compliance Officer only
    @service_auth(cpl_ofcr_only=True)
    def process(node: sy.Device, msg: sy.ReprMessage, verify_key: VerifyKey) -> None:
        pass

    process(node=node, msg=msg, verify_key=node.root_verify_key)



# third party
import tenseal as ts

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...proto.lib.tenseal.vector_pb2 import TenSEALVector as TenSEALVector_PB
from ..util import full_name_with_name


def object2proto(obj: object) -> TenSEALVector_PB:
    proto = TenSEALVector_PB()
    proto.obj_type = full_name_with_name(klass=obj._sy_serializable_wrapper_type)  # type: ignore
    proto.vector = obj.serialize()  # type: ignore

    return proto


def proto2object(proto: TenSEALVector_PB) -> ts.CKKSVector:
    vec = ts.lazy_ckks_vector_from(proto.vector)

    return vec


GenerateWrapper(
    wrapped_type=ts.CKKSVector,
    import_path="tenseal.CKKSVector",
    protobuf_scheme=TenSEALVector_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



"""A Pointer is the main handler when interacting with remote data.
A Pointer object represents an API for interacting with data (of any type)
at a specific location. The pointer should never be instantiated, only subclassed.

The relation between pointers and data is many to one,
there can be multiple pointers pointing to the same piece of data, meanwhile,
a pointer cannot point to multiple data sources.

A pointer is just an object id on a remote location and a set of methods that can be
executed on the remote machine directly on that object. One note that has to be made
is that all operations between pointers will return a pointer, the only way to have access
to the result is by calling .get() on the pointer.

There are two proper ways of receiving a pointer on some data:

1. When sending that data on a remote machine the user receives a pointer.
2. When the user searches for the data in an object store it receives a pointer to that data,
   if it has the correct permissions for that.

After receiving a pointer, one might want to get the data behind the pointer locally. For that the
user should:

1. Request access by calling .request().

Example:

.. code-block::

    pointer_object.request(name = "Request name", reason = "Request reason")

2. The data owner has to approve the request (check the domain node docs).
3. The data user checks if the request has been approved (check the domain node docs).
4. After the request has been approved, the data user can call .get() on the pointer to get the
   data locally.

Example:

.. code-block::

    pointer_object.get()

Pointers are being generated for most types of objects in the data science scene, but what you can
do on them is not the pointers job, see the lib module for more details. One can see the pointer
as a proxy to the actual data, the filtering and the security being applied where the data is being
held.

Example:

.. code-block::

    # creating the data holder domain
    domain_1 = Domain(name="Data holder domain")

    # creating dummy data
    tensor = th.tensor([1, 2, 3])

    # creating the data holder client
    domain_1_client = domain_1.get_root_client()

    # sending the data to the client and receiving a pointer of that data.
    data_ptr_domain_1 = tensor.send(domain_1_client)

    # creating the data user domain
    domain_2 = Domain(name="Data user domain")

    # creating a request to access the data
    data_ptr_domain_1.request(
        name="My Request", reason="I'd lke to see this pointer"
    )

    # getting the remote id of the object
    requested_object = data_ptr_domain_1.id_at_location

    # getting the request id
    message_request_id = domain_1_client.requests.get_request_id_from_object_id(
        object_id=requested_object
    )

    # the data holder accepts the request
    domain_1.requests[0].owner_client_if_available = domain_1_client
    domain_1.requests[0].accept()

    # the data user checks if the data holder approved his request
    response = data_ptr_domain_1.check_access(node=domain_2, request_id=message_request_id)

"""
# stdlib
import time
from typing import Any
from typing import List
from typing import Optional
import warnings

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft absolute
import syft as sy

# syft relative
from ...logger import debug
from ...logger import error
from ...logger import warning
from ...proto.core.pointer.pointer_pb2 import Pointer as Pointer_PB
from ..common.pointer import AbstractPointer
from ..common.serde.deserialize import _deserialize
from ..common.serde.serializable import bind_protobuf
from ..common.uid import UID
from ..io.address import Address
from ..node.abstract.node import AbstractNode
from ..node.common.action.garbage_collect_object_action import (
    GarbageCollectObjectAction,
)
from ..node.common.action.get_object_action import GetObjectAction
from ..node.common.service.get_repr_service import GetReprMessage
from ..node.common.service.obj_search_permission_service import (
    ObjectSearchPermissionUpdateMessage,
)
from ..store.storeable_object import StorableObject


# TODO: Fix the Client, Address, Location confusion
@bind_protobuf
class Pointer(AbstractPointer):
    """
    The pointer is the handler when interacting with remote data.

    Automatically generated subclasses of Pointer need to be able to look up
    the path and name of the object type they point to as a part of serde. For more
    information on how subclasses are automatically generated, please check the ast
    module.

    :param location: The location where the data is being held.
    :type location: Address
    :param id_at_location: The UID of the object on the remote location.
    :type id_at_location: UID
    """

    path_and_name: str
    _pointable: bool = False

    def __init__(
        self,
        client: Any,
        id_at_location: Optional[UID] = None,
        object_type: str = "",
        tags: Optional[List[str]] = None,
        description: str = "",
    ) -> None:
        super().__init__(
            client=client,
            id_at_location=id_at_location,
            tags=tags,
            description=description,
        )
        self.object_type = object_type
        # _exhausted becomes True in get() call
        # when delete_obj is True and network call
        # has already been made
        self._exhausted = False

    def _get(self, delete_obj: bool = True, verbose: bool = False) -> StorableObject:
        """Method to download a remote object from a pointer object if you have the right
        permissions.

        :return: returns the downloaded data
        :rtype: StorableObject
        """

        debug(
            f"> GetObjectAction for id_at_location={self.id_at_location} "
            + f"with delete_obj={delete_obj}"
        )
        obj_msg = GetObjectAction(
            id_at_location=self.id_at_location,
            address=self.client.address,
            reply_to=self.client.address,
            delete_obj=delete_obj,
        )

        obj = self.client.send_immediate_msg_with_reply(msg=obj_msg).data
        if self.is_enum:
            enum_class = self.client.lib_ast.query(self.path_and_name).object_ref
            return enum_class(obj)

        return obj

    def get_copy(
        self,
        request_block: bool = False,
        timeout_secs: int = 20,
        reason: str = "",
        verbose: bool = False,
    ) -> Optional[StorableObject]:
        """Method to download a remote object from a pointer object if you have the right
        permissions. Optionally can block while waiting for approval.

        :return: returns the downloaded data
        :rtype: Optional[StorableObject]
        """
        return self.get(
            request_block=request_block,
            timeout_secs=timeout_secs,
            reason=reason,
            delete_obj=False,
            verbose=verbose,
        )

    def print(self) -> "Pointer":
        obj = None
        try:
            obj_msg = GetReprMessage(
                id_at_location=self.id_at_location,
                address=self.client.address,
                reply_to=self.client.address,
            )

            obj = self.client.send_immediate_msg_with_reply(msg=obj_msg).repr
        except Exception as e:
            if "You do not have permission to .get()" in str(
                e
            ) or "UnknownPrivateException" in str(e):
                # syft relative
                from ..node.domain.service import RequestStatus

                response_status = self.request(
                    reason="Calling remote print",
                    block=True,
                    timeout_secs=3,
                )
                if (
                    response_status is not None
                    and response_status == RequestStatus.Accepted
                ):
                    return self.print()

        # TODO: Create a remote print interface for objects which displays them in a
        # nice way, we could also even buffer this between chained ops until we return
        # so that we can print once and display a nice list of data and ops
        # issue: https://github.com/OpenMined/PySyft/issues/5167
        if obj is not None:
            print(obj)
        else:
            print(f"No permission to print() {self}")

        return self

    def get(
        self,
        request_block: bool = False,
        timeout_secs: int = 20,
        reason: str = "",
        delete_obj: bool = True,
        verbose: bool = False,
    ) -> Optional[StorableObject]:
        """Method to download a remote object from a pointer object if you have the right
        permissions. Optionally can block while waiting for approval.

        :return: returns the downloaded data
        :rtype: Optional[StorableObject]
        """
        # syft relative
        from ..node.domain.service import RequestStatus

        if self._exhausted:
            raise ReferenceError(
                "Object has already been deleted. This pointer is exhausted"
            )

        if not request_block:
            result = self._get(delete_obj=delete_obj, verbose=verbose)
        else:
            response_status = self.request(
                reason=reason,
                block=True,
                timeout_secs=timeout_secs,
                verbose=verbose,
            )
            if (
                response_status is not None
                and response_status == RequestStatus.Accepted
            ):
                result = self._get(delete_obj=delete_obj, verbose=verbose)
            else:
                return None

        if result is not None and delete_obj:
            self.gc_enabled = False
            self._exhausted = True

        return result

    def _object2proto(self) -> Pointer_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: Pointer_PB

        .. note::
            This method is purely an internal method. Please use sy.serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return Pointer_PB(
            points_to_object_with_path=self.path_and_name,
            pointer_name=type(self).__name__,
            id_at_location=sy.serialize(self.id_at_location),
            location=sy.serialize(self.client.address),
            tags=self.tags,
            description=self.description,
            object_type=self.object_type,
        )

    @staticmethod
    def _proto2object(proto: Pointer_PB) -> "Pointer":
        """Creates a Pointer from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of Pointer
        :rtype: Pointer

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """
        # TODO: we need _proto2object to include a reference to the node doing the
        # deserialization so that we can convert location into a client object. At present
        # it is an address object which will cause things to break later.

        points_to_type = sy.lib_ast.query(proto.points_to_object_with_path)
        pointer_type = getattr(points_to_type, proto.pointer_name)
        # WARNING: This is sending a serialized Address back to the constructor
        # which currently depends on a Client for send_immediate_msg_with_reply
        return pointer_type(
            id_at_location=_deserialize(blob=proto.id_at_location),
            client=_deserialize(blob=proto.location),
            tags=proto.tags,
            description=proto.description,
            object_type=proto.object_type,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return Pointer_PB

    def request(
        self,
        reason: str = "",
        block: bool = False,
        timeout_secs: Optional[int] = None,
        verbose: bool = False,
    ) -> Any:
        """Method that requests access to the data on which the pointer points to.

        Example:

        .. code-block::

            # data holder domain
            domain_1 = Domain(name="Data holder")

            # data
            tensor = th.tensor([1, 2, 3])

            # generating the client for the domain
            domain_1_client = domain_1.get_root_client()

            # sending the data and receiving a pointer
            data_ptr_domain_1 = tensor.send(domain_1_client)

            # requesting access to the pointer
            data_ptr_domain_1.request(name="My Request", reason="Research project.")

        :param name: The title of the request that the data owner is going to see.
        :type name: str
        :param reason: The description of the request. This is the reason why you want to have
            access to the data.
        :type reason: str

        .. note::
            This method should be used when the remote data associated with the pointer wants to be
            downloaded locally (or use .get() on the pointer).
        """
        # syft relative
        from ..node.domain.service import RequestMessage

        # if you request non-blocking you don't need a timeout
        # if you request blocking you need a timeout, so lets set a default on here
        # a timeout of 0 would be a way to say don't block my local notebook but if the
        # duet partner has a rule configured it will get executed first before the
        # request would time out
        if timeout_secs is None and block is False:
            timeout_secs = -1  # forever

        msg = RequestMessage(
            request_description=reason,
            address=self.client.address,
            owner_address=self.client.address,
            object_id=self.id_at_location,
            object_type=self.object_type,
            requester_verify_key=self.client.verify_key,
            timeout_secs=timeout_secs,
        )

        self.client.send_immediate_msg_without_reply(msg=msg)

        # wait long enough for it to arrive and trigger a handler
        time.sleep(0.1)

        if not block:
            return None
        else:
            if timeout_secs is None:
                timeout_secs = 30  # default if not explicitly set

            # syft relative
            from ..node.domain.service import RequestAnswerMessage
            from ..node.domain.service import RequestStatus

            output_string = "> Waiting for Blocking Request: "
            output_string += f"  {self.id_at_location}"
            if len(reason) > 0:
                output_string += f": {reason}"
            if len(output_string) > 0 and output_string[-1] != ".":
                output_string += "."
            debug(output_string)
            status = None
            start = time.time()

            last_check: float = 0.0
            while True:
                now = time.time()
                try:
                    # won't run on the first pass because status is None which allows
                    # for remote request handlers to auto respond before timeout
                    if now - start > timeout_secs:
                        log = (
                            f"\n> Blocking Request Timeout after {timeout_secs} seconds"
                        )
                        debug(log)
                        return status

                    # only check once every second
                    if now - last_check > 1:
                        last_check = now
                        debug(f"> Sending another Request Message {now - start}")
                        status_msg = RequestAnswerMessage(
                            request_id=msg.id,
                            address=self.client.address,
                            reply_to=self.client.address,
                        )
                        response = self.client.send_immediate_msg_with_reply(
                            msg=status_msg
                        )
                        status = response.status
                        if response.status == RequestStatus.Pending:
                            time.sleep(0.1)
                            continue
                        else:
                            # accepted or rejected lets exit
                            status_text = "REJECTED"
                            if status == RequestStatus.Accepted:
                                status_text = "ACCEPTED"
                            log = f" {status_text}"
                            debug(log)
                            return status
                except Exception as e:
                    error(f"Exception while running blocking request. {e}")
                    # escape the while loop
                    return status

    @property
    def searchable(self) -> bool:
        msg = "`searchable` is deprecated please use `pointable` in future"
        warning(msg, print=True)
        warnings.warn(
            msg,
            DeprecationWarning,
        )
        return self._pointable

    @searchable.setter
    def searchable(self, value: bool) -> None:
        msg = "`searchable` is deprecated please use `pointable` in future"
        warning(msg, print=True)
        warnings.warn(
            msg,
            DeprecationWarning,
        )
        self.pointable = value

    @property
    def pointable(self) -> bool:
        return self._pointable

    @pointable.setter
    def pointable(self, value: bool) -> None:
        if value != self._pointable:
            self.update_searchability(not self._pointable)

    def update_searchability(
        self,
        pointable: bool = True,
        target_verify_key: Optional[VerifyKey] = None,
        searchable: Optional[bool] = None,
    ) -> None:
        """Make the object pointed at pointable or not for other people. If
        target_verify_key is not specified, the searchability for the VERIFYALL group
        will be toggled.

        :param pointable: If the target object should be made pointable or not.
        :type target_verify_key: bool
        :param target_verify_key: The verify_key of the client to which we want to give
               search permission.
        :type target_verify_key: Optional[VerifyKey]
        """

        if searchable is not None:
            warn_msg = "`searchable` is deprecated please use `pointable` in future"
            warning(warn_msg, print=True)
            warnings.warn(
                warn_msg,
                DeprecationWarning,
            )
            pointable = searchable

        self._pointable = pointable
        msg = ObjectSearchPermissionUpdateMessage(
            add_instead_of_remove=pointable,
            target_verify_key=target_verify_key,
            target_object_id=self.id_at_location,
            address=self.client.address,
        )
        self.client.send_immediate_msg_without_reply(msg=msg)

    def check_access(self, node: AbstractNode, request_id: UID) -> any:  # type: ignore
        """Method that checks the status of an already made request. There are three
        possible outcomes when requesting access:

        1. RequestStatus.Accepted - your request has been approved, you can not .get() your data.
        2. RequestStatus.Pending - your request has not been reviewed yet.
        3. RequestStatus.Rejected - your request has been rejected.

        :param node: The node that queries the request status.
        :type node: AbstractNode
        :param request_id: The request on which you are querying the status.
        :type request_id: UID
        """
        # syft relative
        from ..node.domain.service import RequestAnswerMessage

        msg = RequestAnswerMessage(
            request_id=request_id, address=self.client.address, reply_to=node.address
        )
        response = self.client.send_immediate_msg_with_reply(msg=msg)

        return response.status

    def __del__(self) -> None:
        _client_type = type(self.client)
        if (_client_type == Address) or issubclass(_client_type, AbstractNode):
            # it is a serialized pointer that we receive from another client do nothing
            return
        if self.gc_enabled:
            # Create the delete message
            msg = GarbageCollectObjectAction(
                id_at_location=self.id_at_location, address=self.client.address
            )

            # Send the message
            self.client.send_eventual_msg_without_reply(msg=msg)



# syft relative
from ...core.node.network.client import NetworkClient
from ..connections.http_connection import HTTPConnection
from .signaling_client import SignalingClient

WebRTC_HOST = (
    "http://ec2-18-191-23-46.us-east-2.compute.amazonaws.com:5000"  # noqa: F811
)


def register(
    url: str = WebRTC_HOST,
) -> SignalingClient:

    client = SignalingClient(
        url=url, conn_type=HTTPConnection, client_type=NetworkClient
    )
    return client



# stdlib
import sys
import time


def do_test(port: int) -> None:
    # third party
    import torch

    # syft absolute
    import syft as sy

    sy.logger.add(sys.stderr, "ERROR")

    duet = sy.launch_duet(loopback=True, network_url=f"http://127.0.0.1:{port}/")
    duet.requests.add_handler(action="accept")

    t = torch.randn(20).reshape(4, 5)
    _ = t.send(duet, pointable=True)

    sy.core.common.event_loop.loop.run_forever()


def ds_test(port: int) -> None:
    # third party
    import torch

    # syft absolute
    import syft as sy

    sy.logger.add(sys.stderr, "ERROR")

    duet = sy.join_duet(loopback=True, network_url=f"http://127.0.0.1:{port}/")

    time.sleep(3)

    data = duet.store[0].get(request_block=True, delete_obj=False)

    assert data.shape == torch.Size([4, 5]), data.shape


test_scenario_torch_tensor_sanity = (
    "test_scenario_torch_tensor_sanity",
    do_test,
    ds_test,
)



# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# third party
import tenseal as ts

# syft relative
from . import bfv_vector  # noqa: 401
from . import ckks_tensor  # noqa: 401
from . import ckks_vector  # noqa: 401
from . import context  # noqa: 401
from . import plain_tensor  # noqa: 401
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..util import generic_update_ast

LIB_NAME = "tenseal"
PACKAGE_SUPPORT = {
    "lib": LIB_NAME,
    "python": {"max_version": (3, 9, 99)},
}


def create_ast(client: TypeAny) -> Globals:
    ast = Globals(client=client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [("tenseal", ts)]

    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        ("tenseal.Context", "tenseal.Context", ts.Context),
        ("tenseal.SCHEME_TYPE", "tenseal.SCHEME_TYPE", ts.SCHEME_TYPE),
        ("tenseal.CKKSVector", "tenseal.CKKSVector", ts.CKKSVector),
        ("tenseal.CKKSTensor", "tenseal.CKKSTensor", ts.CKKSTensor),
        ("tenseal.BFVVector", "tenseal.BFVVector", ts.BFVVector),
        ("tenseal.PlainTensor", "tenseal.PlainTensor", ts.PlainTensor),
    ]

    methods = [
        ("tenseal.SCHEME_TYPE.BFV", "tenseal.SCHEME_TYPE"),
        ("tenseal.SCHEME_TYPE.CKKS", "tenseal.SCHEME_TYPE"),
        ("tenseal.SCHEME_TYPE.NONE", "tenseal.SCHEME_TYPE"),
        # Context
        ("tenseal.Context.generate_galois_keys", "syft.lib.python._SyNone"),
        ("tenseal.Context.generate_relin_keys", "syft.lib.python._SyNone"),
        ("tenseal.Context.has_galois_keys", "syft.lib.python.Bool"),
        ("tenseal.Context.has_relin_keys", "syft.lib.python.Bool"),
        ("tenseal.Context.has_public_key", "syft.lib.python.Bool"),
        ("tenseal.Context.has_secret_key", "syft.lib.python.Bool"),
        ("tenseal.Context.is_public", "syft.lib.python.Bool"),
        ("tenseal.Context.is_private", "syft.lib.python.Bool"),
        ("tenseal.Context.make_context_public", "syft.lib.python._SyNone"),
        ("tenseal.Context.copy", "tenseal.Context"),
        ("tenseal.Context.global_scale", "syft.lib.python.Float"),
        ("tenseal.Context.auto_mod_switch", "syft.lib.python.Bool"),
        ("tenseal.Context.auto_relin", "syft.lib.python.Bool"),
        ("tenseal.Context.auto_rescale", "syft.lib.python.Bool"),
        # CKKSVector
        ("tenseal.CKKSVector.__add__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__iadd__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__radd__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__mul__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__imul__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__rmul__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__sub__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__isub__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__rsub__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__pow__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__ipow__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__neg__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.add", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.add_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.mul", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.mul_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.sub", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.sub_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.neg", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.neg_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.sum", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.sum_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.square", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.square_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.pow", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.pow_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.polyval", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.polyval_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.dot", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector._dot", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.dot_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.mm", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector._mm", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.mm_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.matmul", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.matmul_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__matmul__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.__imatmul__", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.conv2d_im2col", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.conv2d_im2col_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector._enc_matmul_plain", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.enc_matmul_plain", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.enc_matmul_plain_", "tenseal.CKKSVector"),
        ("tenseal.CKKSVector.decrypt", "syft.lib.python.List"),
        ("tenseal.CKKSVector.link_context", "syft.lib.python._SyNone"),
        # CKKSTensor
        ("tenseal.CKKSTensor.__add__", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.__iadd__", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.__radd__", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.__mul__", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.__imul__", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.__rmul__", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.__sub__", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.__isub__", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.__rsub__", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.__pow__", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.__ipow__", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.__neg__", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.add", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.add_", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.mul", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.mul_", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.sub", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.sub_", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.neg", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.neg_", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.sum", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.sum_", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.square", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.square_", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.pow", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.pow_", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.polyval", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.polyval_", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.dot", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.dot_", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.shape", "syft.lib.python.List"),
        ("tenseal.CKKSTensor.reshape", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.reshape_", "tenseal.CKKSTensor"),
        ("tenseal.CKKSTensor.decrypt", "syft.lib.python.List"),
        ("tenseal.CKKSTensor.link_context", "syft.lib.python._SyNone"),
        # BFVVector
        ("tenseal.BFVVector.__add__", "tenseal.BFVVector"),
        ("tenseal.BFVVector.__iadd__", "tenseal.BFVVector"),
        ("tenseal.BFVVector.__radd__", "tenseal.BFVVector"),
        ("tenseal.BFVVector.__mul__", "tenseal.BFVVector"),
        ("tenseal.BFVVector.__imul__", "tenseal.BFVVector"),
        ("tenseal.BFVVector.__rmul__", "tenseal.BFVVector"),
        ("tenseal.BFVVector.__sub__", "tenseal.BFVVector"),
        ("tenseal.BFVVector.__isub__", "tenseal.BFVVector"),
        ("tenseal.BFVVector.__rsub__", "tenseal.BFVVector"),
        ("tenseal.BFVVector.add", "tenseal.BFVVector"),
        ("tenseal.BFVVector.add_", "tenseal.BFVVector"),
        ("tenseal.BFVVector.mul", "tenseal.BFVVector"),
        ("tenseal.BFVVector.mul_", "tenseal.BFVVector"),
        ("tenseal.BFVVector.sub", "tenseal.BFVVector"),
        ("tenseal.BFVVector.sub_", "tenseal.BFVVector"),
        ("tenseal.BFVVector.decrypt", "syft.lib.python.List"),
        ("tenseal.BFVVector.link_context", "syft.lib.python._SyNone"),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)



# stdlib
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from syft_proto.execution.v1.placeholder_pb2 import Placeholder as PlaceholderPB
import torch as th

# syft relative
from ...core.common.object import Serializable
from .common import get_protobuf_id
from .common import set_protobuf_id
from .placeholder_id import PlaceholderId


class PlaceHolder(Serializable):
    def __init__(
        self,
        id: Union[int, str],
        tags: Set = set(),
        description: Optional[str] = None,
        shape: Optional[Union[th.Size, Tuple[int]]] = None,
        expected_dtype: Optional[th.dtype] = None,
    ):
        """A PlaceHolder acts as a tensor but does nothing special. It can get
        "instantiated" when a real tensor is appended as a child attribute. It
        will send forward all the commands it receives to its child tensor.

        When you send a PlaceHolder, you don't sent the instantiated tensors.

        Args:
            id: An optional string or integer id of the PlaceHolder.
        """
        super().__init__()

        self.id = PlaceholderId(id)
        self.tags = tags
        self.description = description
        self.expected_shape: Optional[Tuple] = (
            tuple(shape) if shape is not None else None
        )
        self.expected_dtype = expected_dtype
        self.child: Optional[Union[th.Tensor, th.nn.Parameter, "PlaceHolder"]] = None

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """
        return PlaceholderPB

    def _object2proto(self) -> PlaceholderPB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: ObjectWithID_PB

        .. note::
            This method is purely an internal method. Please use object.serialize() or one of
            the other public serialization methods if you wish to serialize an
            object.
        """

        protobuf_placeholder = PlaceholderPB()
        set_protobuf_id(protobuf_placeholder.id, self.id.value)
        protobuf_placeholder.tags.extend(self.tags)

        if self.description:
            protobuf_placeholder.description = self.description

        if self.expected_shape:
            protobuf_placeholder.expected_shape.dims.extend(self.expected_shape)

        return protobuf_placeholder

    @staticmethod
    def _proto2object(proto: PlaceholderPB) -> "PlaceHolder":
        """Creates a ObjectWithID from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of Plan
        :rtype: Plan

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        tensor_id = get_protobuf_id(proto.id)
        tags = set(proto.tags) if proto.tags else set()

        description = None
        if bool(proto.description):
            description = proto.description

        expected_shape = tuple(proto.expected_shape.dims) or None

        return PlaceHolder(
            id=tensor_id, tags=tags, description=description, shape=expected_shape
        )

    def instantiate(self, tensor: Union[th.Tensor, th.nn.Parameter]) -> "PlaceHolder":
        """
        Add a tensor as a child attribute. All operations on the placeholder will be also
        executed on this child tensor.

        We remove Placeholders if is there are any.
        """
        if isinstance(tensor, PlaceHolder):
            self.child = tensor.child
        else:
            self.child = tensor

        shape = getattr(self.child, "shape", None)
        if shape is not None:
            self.expected_shape = tuple(shape)

        dtype = getattr(self.child, "dtype", None)
        if dtype is not None:
            self.expected_dtype = dtype

        return self



# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey
from typing_extensions import final

# syft relative
from ..... import serialize
from .....core.common.serde.serializable import bind_protobuf
from .....proto.core.node.common.service.get_repr_service_pb2 import (
    GetReprMessage as GetReprMessage_PB,
)
from .....proto.core.node.common.service.get_repr_service_pb2 import (
    GetReprReplyMessage as GetReprReplyMessage_PB,
)
from .....util import traceback_and_raise
from ....common.group import VERIFYALL
from ....common.message import ImmediateSyftMessageWithReply
from ....common.message import ImmediateSyftMessageWithoutReply
from ....common.serde.deserialize import _deserialize
from ....common.uid import UID
from ....io.address import Address
from ...abstract.node import AbstractNode
from .auth import service_auth
from .node_service import ImmediateNodeServiceWithReply


@bind_protobuf
@final
class GetReprMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self,
        address: Address,
        reply_to: Address,
        id_at_location: UID,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)
        self.id_at_location = id_at_location

    def _object2proto(self) -> GetReprMessage_PB:
        return GetReprMessage_PB(
            id_at_location=serialize(self.id_at_location),
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(proto: GetReprMessage_PB) -> "GetReprMessage":
        return GetReprMessage(
            id_at_location=_deserialize(blob=proto.id_at_location),
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
            reply_to=_deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return GetReprMessage_PB


@bind_protobuf
class GetReprReplyMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        repr: str,
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.repr = repr

    def _object2proto(self) -> GetReprReplyMessage_PB:
        return GetReprReplyMessage_PB(
            repr=self.repr,
            msg_id=serialize(self.id),
            address=serialize(self.address),
        )

    @staticmethod
    def _proto2object(proto: GetReprReplyMessage_PB) -> "GetReprReplyMessage":
        return GetReprReplyMessage(
            repr=proto.repr,
            msg_id=_deserialize(blob=proto.msg_id),
            address=_deserialize(blob=proto.address),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return GetReprReplyMessage_PB


class GetReprService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(root_only=True)
    def process(
        node: AbstractNode,
        msg: GetReprMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> GetReprReplyMessage:
        if verify_key is None:
            traceback_and_raise(
                "Can't process an GetReprService with no verification key."
            )

        obj = node.store[msg.id_at_location]
        contains_all_in_permissions = any(
            key is VERIFYALL for key in obj.read_permissions.keys()
        )

        if not (
            verify_key in obj.read_permissions.keys()
            or verify_key == node.root_verify_key
            or contains_all_in_permissions
        ):
            raise PermissionError("Permission to get repr of object not granted!")
        else:
            # TODO: Create a remote print interface for objects which displays them in a
            # nice way, we could also even buffer this between chained ops until we
            # return so that we can print once and display a nice list of data and ops
            # issue: https://github.com/OpenMined/PySyft/issues/5167
            result = repr(obj.data)
            return GetReprReplyMessage(repr=result, address=msg.reply_to)

    @staticmethod
    def message_handler_types() -> List[Type[GetReprMessage]]:
        return [GetReprMessage]



# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# third party
import PIL

# syft relative
from . import image  # noqa: 401
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..util import generic_update_ast

LIB_NAME = "PIL"
PACKAGE_SUPPORT = {"lib": LIB_NAME}


def create_ast(client: TypeAny = None) -> Globals:
    ast = Globals(client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [
        ("PIL", PIL),
        ("PIL.Image", PIL.Image),
    ]

    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        ("PIL.Image.Image", "PIL.Image.Image", PIL.Image.Image)
    ]

    methods: TypeList[TypeTuple[str, str]] = []

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)
